"""Metadata extraction example with rasterio."""

from pystac_client import Client
import rasterio.features  # type: ignore
import rasterio.warp  # type: ignore
from teledetection import sign_inplace

api = Client.open(
    "https://api.stac.teledetection.fr",
    modifier=sign_inplace,
)

YEAR = 2022
bbox = [4, 42.99, 5, 44.05]
res = api.search(bbox=bbox, datetime=[f"{YEAR}-01-01", f"{YEAR}-12-25"])

for item in res.items():
    url = item.assets["src_xs"].href
    with rasterio.open(url) as dataset:
        # Read the dataset's valid data mask as a ndarray.
        mask = dataset.dataset_mask()

        # Extract feature shapes and values from the array.
        for geom, val in rasterio.features.shapes(mask, transform=dataset.transform):
            # Transform shapes from the dataset's own coordinate
            # reference system to CRS84 (EPSG:4326).
            geom = rasterio.warp.transform_geom(
                dataset.crs, "EPSG:4326", geom, precision=6
            )

            # Print GeoJSON shapes to stdout.
            print(geom)



"""Teledetection SDK module."""

# flake8: noqa

from importlib.metadata import version, PackageNotFoundError
from teledetection.sdk.signing import (
    sign,
    sign_inplace,
    sign_urls,
    sign_item,
    sign_asset,
    sign_item_collection,
    sign_url_put,
)  # noqa
from .sdk.oauth2 import OAuth2Session  # noqa
from .sdk.http import get_headers, get_userinfo, get_username

try:
    __version__ = version("teledetection")
except PackageNotFoundError:
    pass



"""This module is used to upload files using HTTP requests."""

from teledetection.sdk.signing import sign_url_put
from teledetection.sdk.utils import create_session


def push(local_filename: str, target_url: str):
    """Publish a local file to the cloud."""
    remote_presigned_url = sign_url_put(target_url)

    session = create_session()

    with open(local_filename, "rb") as f:
        ret = session.put(remote_presigned_url, data=f, timeout=10)

    if ret.status_code == 200:
        return remote_presigned_url

    ret.raise_for_status()
    return ""



"""TOA images mosaicing with pyotb."""

# pylint: disable=duplicate-code

from pystac_client import Client
import pyotb  # type: ignore
from teledetection import sign_inplace

api = Client.open(
    "https://api.stac.teledetection.fr",
    modifier=sign_inplace,
)

res = api.search(
    bbox=[4, 42.99, 5, 44.05],
    datetime=["2022-01-01", "2022-12-25"],
    collections=["spot-6-7-drs"],
)

urls = [f"/vsicurl/{r.assets['src_xs'].href}" for r in res.items()]
toa_images = [pyotb.OpticalCalibration({"in": url}) for url in urls]
mosa = pyotb.Mosaic({"il": toa_images})
mosa.write("toa_mosa.tif?&box=5000:5000:4096:4096")



"""Module dedicated to OAuth2 device flow."""

import datetime
import io
import time
from abc import abstractmethod
from typing import Dict
from urllib.parse import urljoin
import qrcode

from .logger import get_logger_for  # type: ignore
from .utils import create_session
from .model import JWT, DeviceGrantResponse
from .settings import ENV

log = get_logger_for(__name__)
TIMEOUT = ENV.tld_request_timeout


def retrieve_token_endpoint():
    """Retrieve the token endpoint from the s3 signing endpoint."""
    openapi_url = urljoin(ENV.tld_signing_endpoint, "openapi.json")
    log.debug("Fetching OAuth2 endpoint from openapi url %s", openapi_url)
    _session = create_session()
    res = _session.get(openapi_url, timeout=TIMEOUT)
    res.raise_for_status()
    data = res.json()
    return data["components"]["securitySchemes"]["OAuth2PasswordBearer"]["flows"][
        "password"
    ]["tokenUrl"]


class RefreshTokenError(Exception):
    """Unable to refresh token."""


class ExpiredAuthLinkError(Exception):
    """Authentication link has expired."""


class GrantMethodBase:
    """Base class for grant methods."""

    client_id: str
    headers: Dict[str, str] = {"Content-Type": "application/x-www-form-urlencoded"}
    _token_endpoint: str | None = None

    def get_token_endpoint(self):
        """Get the token endpoint."""
        if not self._token_endpoint:
            self._token_endpoint = retrieve_token_endpoint()
        return retrieve_token_endpoint()

    @abstractmethod
    def get_first_token(self) -> JWT:
        """Provide the first used token."""
        raise NotImplementedError

    @property
    def data_base(self) -> Dict[str, str]:
        """Base payload."""
        return {
            "client_id": self.client_id,
            "scope": "openid offline_access",
        }

    def refresh_token(self, old_jwt: JWT) -> JWT:
        """Refresh the token."""
        log.debug("Refreshing token")
        assert old_jwt, "JWT is empty"
        data = self.data_base.copy()
        data.update(
            {
                "refresh_token": old_jwt.refresh_token,
                "grant_type": "refresh_token",
            }
        )
        ret = create_session().post(
            self.get_token_endpoint(),
            headers=self.headers,
            data=data,
            timeout=TIMEOUT,
        )
        if ret.status_code == 200:
            log.debug(ret.text)
            return JWT.from_dict(ret.json())
        raise RefreshTokenError


class DeviceGrant(GrantMethodBase):
    """Device grant method."""

    client_id = "gdal"

    def get_first_token(self) -> JWT:
        """Get the first JWT token."""
        device_endpoint = f"{self.get_token_endpoint().rsplit('/', 1)[0]}/auth/device"

        req = create_session()
        log.debug("Getting token using device authorization grant")
        ret = req.post(
            device_endpoint,
            headers=self.headers,
            data=self.data_base,
            timeout=TIMEOUT,
        )
        if ret.status_code == 200:
            response = DeviceGrantResponse(**ret.json())
            verif_url_comp = response.verification_uri_complete
            log.info("Open the following URL in your browser to grant access:")
            log.info("\033[92m %s \033[0m", verif_url_comp)

            # QR code
            qr_code = qrcode.QRCode()
            qr_code.add_data(verif_url_comp)
            buffer = io.StringIO()
            qr_code.print_ascii(out=buffer)
            buffer.seek(0)
            log.info(buffer.read())

            log.info("Waiting for authentication...")
            start = time.time()
            data = self.data_base.copy()
            grant_type = "urn:ietf:params:oauth:grant-type:device_code"
            data.update(
                {
                    "device_code": response.device_code,
                    "grant_type": grant_type,
                }
            )
            while True:
                ret = req.post(
                    self.get_token_endpoint(),
                    headers=self.headers,
                    data=data,
                    timeout=TIMEOUT,
                )
                elapsed = time.time() - start
                log.debug(
                    "Elapsed: %.0f, status: %s (%.0f seconds left)",
                    elapsed,
                    ret.status_code,
                    response.expires_in - elapsed,
                )
                if elapsed > response.expires_in:
                    raise ExpiredAuthLinkError
                if ret.status_code != 200:
                    time.sleep(response.interval)
                else:
                    return JWT.from_dict(ret.json())
        raise ConnectionError("Unable to authenticate with the SSO")


class OAuth2Session:
    """Class to start an OAuth2 session."""

    def __init__(self, grant_type: type[GrantMethodBase] = DeviceGrant):
        """Initialize."""
        self.grant = grant_type()
        self.jwt_ttl_margin_seconds = 60
        self.jwt_issuance = datetime.datetime(year=1, month=1, day=1)
        self.jwt: JWT | None = None

    def save_token(self, now: datetime.datetime):
        """Save the JWT to disk."""
        if self.jwt:
            self.jwt_issuance = now
            self.jwt.to_config_dir()
        else:
            log.warning("No OAuth2 credentials to save")

    def _init_jwt(self):
        """Initialize the JWT."""
        if not self.jwt:
            # First JWT initialisation
            self.jwt = JWT.from_config_dir()
        if not self.jwt:
            # When JWT is still `None`, we use the grant method
            self.jwt = self.grant.get_first_token()
            self.save_token(datetime.datetime.now())

    def _refresh_if_needed(self):
        """Refresh the token if ttl is too short."""
        self._init_jwt()
        ttl_margin_seconds = 30
        now = datetime.datetime.now()
        jwt_expires_in = datetime.timedelta(seconds=self.jwt.expires_in)
        access_token_ttl = self.jwt_issuance + jwt_expires_in - now
        access_token_ttl_seconds = access_token_ttl.total_seconds()
        log.debug("access_token_ttl is %s", access_token_ttl_seconds)
        if access_token_ttl_seconds < ttl_margin_seconds:
            # Access token in not valid, but refresh might be
            try:
                self.jwt = self.grant.refresh_token(self.jwt)
            except RefreshTokenError as con_err:
                log.warning(
                    "Unable to refresh token (reason: %s). "
                    "Renewing initial authentication.",
                    con_err,
                )
                self.jwt = self.grant.get_first_token()
        else:
            # Token is still valid
            log.debug("Credentials still valid")
            return
        self.save_token(now)

    def get_access_token(self) -> str:
        """Return the access token."""
        self._init_jwt()
        self._refresh_if_needed()
        assert self.jwt
        return self.jwt.access_token



"""Teledetection package Command Line Interface."""

import os
import tempfile
import subprocess
import getpass
from typing import Dict, List
import datetime
import click

from .sdk.logger import get_logger_for

from .sdk.model import ApiKey
from .sdk.http import OAuth2ConnectionMethod
from .sdk.utils import create_session


@click.group(
    help="Teledetection CLI",
    context_settings={
        "help_option_names": ["-h", "--help"],
        "max_content_width": 120,
    },
)
def tld() -> None:
    """Teledetection Command Line Interface."""


log = get_logger_for(__name__)
conn = OAuth2ConnectionMethod()


def _http(route: str, params: dict | None = None):
    """Perform an HTTP request."""
    session = create_session()
    ret = session.get(
        f"{conn.endpoint}{route}",
        timeout=5,
        params=params,
        headers=conn.get_headers(),
    )
    ret.raise_for_status()
    return ret


def _get_all_keys() -> List[str]:
    """Retrieve all API keys."""
    return _http("list_api_keys_with_metadata").json()


def _create_new_key(description: str) -> Dict[str, str]:
    """Create a new API key."""

    def _default_desc():
        """Default description."""
        return (
            f"Created by {getpass.getuser()} on "
            f"{datetime.datetime.now().strftime('%Y-%m-%d %H:%M')}"
        )

    return _http(
        "create_api_key", params={"description": description or _default_desc()}
    ).json()


def do_create_key(description: str):
    """Create a new API key."""
    log.info("New API key %s created", _create_new_key(description=description))


def do_list_keys():
    """List all generated API keys."""
    keys = _get_all_keys()
    if keys:
        log.info("All existing API keys:")
        log.info("Creation date      \tAccess key      \t[Description]")
        # Prints: 2025-05-06 09:22:50  zPL9GaQrokbMCQGe
        for key in keys:
            log.info(
                "%s\t%s\t%s",
                key["created"].split(".")[0],
                key["access-key"],
                key["description"],
            )
    else:
        log.info("No API key found.")


def do_revoke_key(access_key: str):
    """Revoke an API key."""
    _http(f"revoke_api_key?access_key={access_key}")
    log.info(f"API key {access_key} revoked")


def do_revoke_all_keys():
    """Revoke all API keys."""
    keys = _get_all_keys()
    for key in keys:
        do_revoke_key(key["access-key"])
    if not keys:
        log.info("No API key to revoke.")


def do_register_key(description: str):
    """Create and store a new API key."""
    new_key = _create_new_key(description=description)
    ApiKey.from_dict(new_key).to_config_dir()
    log.info("New API key %s created and stored in config directory", new_key)


def do_remove_key(dont_revoke: bool):
    """Delete the stored API key."""
    if not dont_revoke:
        do_revoke_key(ApiKey.from_config_dir().access_key)
    ApiKey.delete_from_config_dir()


API_KEY_OPS = {
    "create": lambda arg: do_create_key(description=arg),
    "revoke": lambda arg: do_revoke_key(access_key=arg),
    "revoke-all": lambda arg: do_revoke_all_keys(),
    "register": lambda arg: do_register_key(description=arg),
    "remove": lambda arg: do_remove_key(dont_revoke=any(arg)),
    "list": lambda arg: do_list_keys(),
}


@tld.command()
@click.argument(
    "operation",
    type=click.Choice(list(API_KEY_OPS.keys()), case_sensitive=False),
    required=False,
)
@click.argument("argument", default="")
@click.pass_context
def apikey(ctx, operation: str, argument: str):
    """Manage API keys.

    \b
    Other operations:
      list        : List all available API keys

    \b
    Operations on locally-stored key:
      register    : Register a new key
      remove      : Remove a key provided in argument

    \b
    Manually input keys:
      create      : Create a new API key
      revoke      : Revoke a specific key
      revoke-all  : Revoke all keys
    """
    if not operation:
        click.echo(ctx.get_help())
        ctx.exit(0)
    API_KEY_OPS[operation](argument)


try:
    from .upload import diff
    from .upload.stac import (
        StacTransactionsHandler,
        StacUploadTransactionsHandler,
        DEFAULT_S3_EP,
        DEFAULT_STAC_EP,
        DEFAULT_S3_STORAGE,
    )

    @tld.command()
    @click.argument("stac_obj_path")
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option(
        "--storage_endpoint",
        type=str,
        help="Storage endpoint assets will be sent to",
        default=DEFAULT_S3_EP,
    )
    @click.option(
        "-b",
        "--storage_bucket",
        help="Storage bucket assets will be sent to",
        type=str,
        default=DEFAULT_S3_STORAGE,
    )
    @click.option(
        "-o",
        "--overwrite",
        is_flag=True,
        default=False,
        help="Overwrite assets if already existing",
    )
    @click.option(
        "--keep_cog_dir",
        help="Set a directory to keep converted COG files",
        type=str,
        nargs=1,
        default="",
    )
    def publish(
        stac_obj_path: str,
        stac_endpoint: str,
        storage_endpoint: str,
        storage_bucket: str,
        overwrite: bool,
        keep_cog_dir: str,
    ):
        """Publish a STAC object (collection or item collection)."""
        StacUploadTransactionsHandler(
            stac_endpoint=stac_endpoint,
            sign=False,
            storage_endpoint=storage_endpoint,
            storage_bucket=storage_bucket,
            assets_overwrite=overwrite,
            keep_cog_dir=keep_cog_dir,
        ).load_and_publish(stac_obj_path)

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option("-c", "--col_id", type=str, help="STAC collection ID", required=True)
    @click.option("-i", "--item_id", type=str, default=None, help="STAC item ID")
    @click.option("-s", "--sign", is_flag=True, default=False, help="Sign assets HREFs")
    @click.option(
        "-p", "--pretty", is_flag=True, default=False, help="Pretty indent JSON"
    )
    @click.option("-o", "--out_json", type=str, help="Output .json file", required=True)
    def grab(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        stac_endpoint: str,
        col_id: str,
        item_id: str,
        sign: bool,
        pretty: bool,
        out_json: str,
    ):
        """Grab a STAC object (collection, or item) and save it as .json."""
        StacTransactionsHandler(stac_endpoint=stac_endpoint, sign=sign).load_and_save(
            col_id=col_id, obj_pth=out_json, item_id=item_id, pretty=pretty
        )

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option("-c", "--col_id", type=str, help="STAC collection ID", required=True)
    @click.option("-i", "--item_id", type=str, default=None, help="STAC item ID")
    def edit(stac_endpoint: str, col_id: str, item_id: str):
        """Edit a STAC object (collection, or item)."""
        with tempfile.NamedTemporaryFile(suffix=".json") as tf:
            StacTransactionsHandler(
                stac_endpoint=stac_endpoint, sign=False
            ).load_and_save(
                col_id=col_id, obj_pth=tf.name, item_id=item_id, pretty=True
            )
            editor = os.environ.get("EDITOR") or "vi"
            subprocess.run([editor, tf.name], check=False)
            StacTransactionsHandler(
                stac_endpoint=stac_endpoint, sign=False
            ).load_and_publish(obj_pth=tf.name)

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option("-c", "--col_id", type=str, help="STAC collection ID", required=True)
    @click.option("-i", "--item_id", type=str, default=None, help="STAC item ID")
    def delete(
        stac_endpoint: str,
        col_id: str,
        item_id: str,
    ):
        """Delete a STAC object (collection or item)."""
        StacTransactionsHandler(
            stac_endpoint=stac_endpoint, sign=False
        ).delete_item_or_col(col_id=col_id, item_id=item_id)

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    def list_cols(
        stac_endpoint: str,
    ):
        """List collections."""
        cols = list(
            StacTransactionsHandler(
                stac_endpoint=stac_endpoint, sign=False
            ).client.get_collections()
        )
        print(f"Found {len(cols)} collection(s):")
        for col in sorted(cols, key=lambda x: x.id):
            print(f"\t{col.id}")

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option("-c", "--col_id", type=str, help="STAC collection ID", required=True)
    @click.option(
        "-m", "--max_items", type=int, help="Max number of items to display", default=20
    )
    @click.option("-s", "--sign", is_flag=True, default=False, help="Sign assets HREFs")
    def list_col_items(stac_endpoint: str, col_id: str, max_items: int, sign: bool):
        """List collection items."""
        items = StacTransactionsHandler(
            stac_endpoint=stac_endpoint, sign=sign
        ).get_items(col_id=col_id, max_items=max_items)
        print(f"Found {len(items)} item(s):")
        for item in items:
            print(f"\t{item.id}")

    @tld.command()
    @click.option(
        "--stac_endpoint",
        help="Endpoint to which STAC objects will be sent",
        type=str,
        default=DEFAULT_STAC_EP,
    )
    @click.option(
        "-p", "--col_path", type=str, help="Local collection path", required=True
    )
    @click.option(
        "-r",
        "--remote_id",
        type=str,
        help="Remote collection ID. If not specified, will use local collection ID",
        required=False,
    )
    def collection_diff(
        stac_endpoint: str,
        col_path: str,
        remote_id: str = "",
    ):
        """List collection items."""
        diff.compare_local_and_upstream(
            StacTransactionsHandler(stac_endpoint=stac_endpoint, sign=False),
            col_path,
            remote_id,
        )
except ImportError:
    log.info(
        "Running CLI without upload support. To install it, use `pip install teledetection[upload]`"
    )



"""NDVI loss example with pyotb."""

from pystac_client import Client
import pyotb  # type: ignore
from teledetection import sign_inplace

api = Client.open(
    "https://api.stac.teledetection.fr",
    modifier=sign_inplace,
)


def mosa(year):
    """Return a pyotb application that perform a mosaic."""
    res = api.search(
        bbox=[4, 42.99, 5, 44.05],
        datetime=[f"{year}-01-01", f"{year}-12-25"],
        collections=["spot-6-7-drs"],
    )
    urls = [f"/vsicurl/{r.assets['src_xs'].href}" for r in res.items()]
    return pyotb.Mosaic({"il": urls})


def ndvi(xs):
    """Return a pyotb application that computes NDVI."""
    return pyotb.BandMath({"il": [xs], "exp": "(im1b4-im1b1)/(im1b4+im1b1)"})


ndvi_22 = ndvi(mosa("2022"))
ndvi_21 = ndvi(mosa("2021"))
delta_ndvi = ndvi_22 - pyotb.Superimpose({"inr": ndvi_22, "inm": ndvi_21})
delta_ndvi.write("ndvi_loss.tif?&box=5000:5000:4096:4096")



"""Some helpers."""

import requests
import urllib3.util.retry

from .settings import ENV


def create_session():
    """Create a session for requests."""
    session = requests.Session()
    retry = urllib3.util.retry.Retry(
        total=ENV.tld_retry_total,
        backoff_factor=ENV.tld_retry_backoff_factor,
        status_forcelist=[429, 500, 502, 503, 504],
    )
    adapter = requests.adapters.HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session



"""Generate the code reference pages."""

from pathlib import Path

import mkdocs_gen_files

for path in sorted(Path("teledetection").rglob("*.py")):  #
    module_path = path.relative_to(".").with_suffix("")  #
    doc_path = path.relative_to(".").with_suffix(".md")  #
    full_doc_path = Path("reference", doc_path)  #

    parts = list(module_path.parts)

    if parts[-1] == "__init__":  #
        parts = parts[:-1]
    elif parts[-1] == "__main__":
        continue

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:  #
        identifier = ".".join(parts)  #
        print("::: " + identifier)
        print("::: " + identifier, file=fd)  #

    mkdocs_gen_files.set_edit_path(full_doc_path, path)



"""Models."""

import os
import json
from typing import Dict
from pydantic import BaseModel, Field, ConfigDict  # pylint: disable = no-name-in-module
from .logger import get_logger_for
from .settings import Settings, get_config_path

log = get_logger_for(__name__)


class Serializable(BaseModel):  # pylint: disable = R0903
    """Base class for serializable pyantic models."""

    model_config = ConfigDict(
        populate_by_name=True,
    )

    @classmethod
    def get_cfg_file_name(cls) -> str | None:
        """Get the config file name (without full path)."""
        name = f".{cls.__name__.lower()}"
        log.debug("Looking for config file for %s", name)
        cfg_pth = get_config_path()
        cfg_file = os.path.join(cfg_pth, name) if cfg_pth else None
        log.debug("Config file %sfound %s", "" if cfg_file else "not ", cfg_file or "")
        return cfg_file

    @classmethod
    def from_config_dir(cls):
        """Try to load from config directory."""
        cfg_file = cls.get_cfg_file_name()
        return cls.from_file(cfg_file) if cfg_file else None

    def to_config_dir(self):
        """Try to save to config files."""
        cfg_file = self.get_cfg_file_name()
        if cfg_file:
            self.to_file(cfg_file)

    @classmethod
    def from_dict(cls, dict: Dict):
        """Get the object from dict."""
        return cls(**dict)

    def to_dict(self) -> Dict[str, str]:
        """To dict."""
        return self.model_dump(by_alias=True)

    @classmethod
    def from_file(cls, file_path: str):
        """Load object from a file."""
        try:
            log.debug("Reading JSON file %s", file_path)
            with open(file_path, "r", encoding="utf-8") as file_handler:
                return cls(**json.load(file_handler))
        except (FileNotFoundError, IOError, json.decoder.JSONDecodeError) as err:
            log.debug("Cannot read object from config directory (%s).", err)

        return None

    def to_file(self, file_path: str):
        """Save the object to file."""
        try:
            log.debug("Writing JSON file %s", file_path)
            with open(file_path, "w", encoding="utf-8") as file_handler:
                json.dump(self.to_dict(), file_handler)
        except IOError as io_err:
            log.warning("Unable to save file %s (%s)", file_path, io_err)

    @classmethod
    def delete_from_config_dir(cls):
        """Delete the config file, if there."""
        cfg_file = cls.get_cfg_file_name()
        if cfg_file:
            os.remove(cfg_file)


class JWT(Serializable):
    """JWT model."""

    access_token: str
    expires_in: int
    refresh_token: str
    refresh_expires_in: int
    token_type: str


class DeviceGrantResponse(BaseModel):  # pylint: disable = R0903
    """Device grant login response model."""

    verification_uri_complete: str
    device_code: str
    expires_in: int
    interval: int


class ApiKey(Serializable):
    """API key class."""

    access_key: str = Field(alias="access-key")
    secret_key: str = Field(alias="secret-key")

    @classmethod
    def from_env(cls):
        """Try to load from env."""
        env = Settings()
        if env.tld_access_key and env.tld_secret_key:
            return cls(
                access_key=env.tld_access_key,
                secret_key=env.tld_secret_key,
            )
        return None

    @classmethod
    def grab(cls):
        """Try to load an API key from env. or file."""
        return cls.from_env() or cls.from_config_dir()



"""Signing module.

Revamp of Microsoft Planetary Computer SAS, using custom URL signing
endpoint instead.
"""

import collections.abc
import math
import re
import time
from copy import deepcopy
from datetime import datetime, timezone
from functools import singledispatch
from typing import Any, Dict, List, Mapping, TypeVar, cast
from enum import Enum
from urllib.parse import parse_qs, urlparse

import pystac_client
from pydantic import BaseModel, ConfigDict  # pylint: disable = no-name-in-module
from pystac import (
    Asset,
    Collection,
    Item,
    ItemCollection,
    STACObjectType,
)
from pystac.serialization.identify import identify_stac_object_type
from pystac_client import ItemSearch

from .http import session
from .settings import S3_STORAGE_DOMAIN, MAX_URLS, ENV
from .logger import get_logger_for


AssetLike = TypeVar("AssetLike", Asset, Dict[str, Any])

asset_xpr = re.compile(
    r"https:\/\/(?P<account>[A-z0-9-.]+?)"
    r"\.meso\.umontpellier\.fr\/(?P<blob>[^<]+)"  # ignore
)

log = get_logger_for(__name__)


class URLBase(BaseModel):  # pylint: disable = R0903
    """Base model for responses."""

    model_config = ConfigDict(populate_by_name=True)
    expiry: datetime


class SignedURL(URLBase):  # pylint: disable = R0903
    """Signed URL response."""

    href: str

    def ttl(self) -> float:
        """Return the number of seconds the token is still valid for."""
        return (self.expiry - datetime.now(timezone.utc)).total_seconds()


class SignedURLBatch(URLBase):  # pylint: disable = R0903
    """Signed URLs (batch of URLs) response."""

    hrefs: dict


# Cache of signing requests so we can reuse them
# Key is the signing URL, value is the S3 token
CACHE: Dict[str, SignedURL] = {}


@singledispatch
def sign(obj: Any, copy: bool = True) -> Any:
    """Sign the relevant URL with a S3 token allowing read access.

    All URLs belonging to supported objects are modified in-place, or returned
    by the function, depending on `copy`.
    Args:
        obj (Any): The object to sign. Must be one of:
            str (URL), Asset, Item, ItemCollection, or ItemSearch, or a
            mapping.
        copy (bool): Whether to sign the object in place, or make a copy.
            Has no effect for immutable objects like strings.

    Returns:
        Any: A copy of the object where all relevant URLs have been signed

    """
    raise TypeError(
        "Invalid type, must be one of: str, Asset, Item, ItemCollection, ItemSearch, or mapping"
    )


def sign_inplace(obj: Any) -> Any:
    """Sign the object in place.

    See :func:`teledetection.sign` for more.

    """
    return sign(obj, copy=False)


def is_vrt_string(string: str) -> bool:
    """Check whether a string looks like a VRT."""
    return string.strip().startswith("<VRTDataset") and string.strip().endswith(
        "</VRTDataset>"
    )


@sign.register(str)
def sign_string(url: str, copy: bool = True) -> str:
    """Sign a URL or VRT-like string containing URLs with a S3 Token.

    Signing with a S3 token allows read access to files in blob storage.

    Args:
        url (str): The HREF of the asset as a URL or a GDAL VRT

            Single URLs can be found on a STAC Item's Asset ``href`` value.
            Only URLs to assets in S3 Storage are signed, other URLs are
            returned unmodified.

            GDAL VRTs can combine many data sources into a single mosaic.
            A VRT can be  built quickly from the GDAL STACIT driver
            https://gdal.org/drivers/raster/stacit.html. Each URL to S3 Storage
            within the VRT is signed.
        copy (bool): No effect.

    Returns:
        str: The signed HREF or VRT

    """
    if is_vrt_string(url):
        return sign_vrt_string(url)
    return sign_urls(urls=[url])[url]


class SignURLRoute(Enum):
    """Different routes used for sign_urls."""

    SIGN_URLS_GET = "sign_urls"
    SIGN_URLS_PUT = "sign_urls_put"


def _generic_sign_urls(urls: List[str], route: SignURLRoute) -> Dict[str, str]:
    """Sign URLs with a S3 Token.

    Signing URL allows read access to files in storage.

    Args:
        urls: List of HREF to sign

            Single URLs can be found on a STAC Item's Asset ``href`` value.
            Only URLs to assets in S3 Storage are signed, other URLs are
            returned unmodified.
        route: API route

    Returns:
        dict of signed HREF: key = original URL, value = signed URL

    """
    signed_urls = {}
    for url in urls:
        stripped_url = url.rstrip("/")
        parsed_url = urlparse(stripped_url)
        if not parsed_url.netloc.endswith(S3_STORAGE_DOMAIN):
            # Outside our domain
            signed_urls[url] = url
        # elif parsed_url.netloc == "????":
        #     # special case for public assets storing thumbnails...
        #     return url
        else:
            parsed_qs = parse_qs(parsed_url.query)
            if set(parsed_qs) & {
                "X-Amz-Security-Token",
                "X-Amz-Signature",
                "X-Amz-Credential",
            }:
                #  looks like we've already signed it
                signed_urls[url] = url

    not_signed_urls = [url for url in urls if url not in signed_urls]
    signed_urls.update(
        {
            url: signed_url.href
            for url, signed_url in _generic_get_signed_urls(
                urls=not_signed_urls, route=route
            ).items()
        }
    )
    return signed_urls


def sign_urls(urls: List[str]) -> Dict[str, str]:
    """Sign multiple URLs for GET."""
    return _generic_sign_urls(urls=urls, route=SignURLRoute(SignURLRoute.SIGN_URLS_GET))


def sign_urls_put(urls: List[str]) -> Dict[str, str]:
    """Sign multiple URLs for PUT."""
    return _generic_sign_urls(urls=urls, route=SignURLRoute(SignURLRoute.SIGN_URLS_PUT))


def sign_url_put(url: str) -> str:
    """Sign a single URL for PUT."""
    urls = sign_urls_put([url])
    return urls[url]


def sign_vrt_string(vrt: str, copy: bool = True) -> str:  # pylint: disable = W0613  # noqa: E501
    """Sign a VRT-like string containing URLs from the storage.

    Signing URLs allows read access to files in storage.

    Args:
        vrt (str): The GDAL VRT

            GDAL VRTs can combine many data sources into a single mosaic. A VRT
            can be built quickly from the GDAL STACIT driver
            https://gdal.org/drivers/raster/stacit.html. Each URL to S3 Storage
            within the VRT is signed.
        copy (bool): No effect.

    Returns:
        str: The signed VRT

    """
    urls = []

    def _repl_vrt(m: re.Match):
        urls.append(m.string[slice(*m.span())])

    asset_xpr.sub(_repl_vrt, vrt)
    signed_urls = sign_urls(urls)

    # The "&" needs to be encoded in signed URLs inside the .vrt
    for url, signed_url in signed_urls.items():
        signed_urls[url] = signed_url.replace("&", "&Amp;")

    # Replace urls with signed urls (single pass)
    rep = dict((re.escape(k), v) for k, v in signed_urls.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], vrt)


@sign.register(Item)
def sign_item(item: Item, copy: bool = True) -> Item:
    """Sign all assets within a PySTAC item.

    Args:
        item (Item): The Item whose assets that will be signed
        copy (bool): Whether to copy (clone) the item or mutate it inplace.

    Returns:
        Item: An Item where all assets' HREFs have
        been replaced with a signed version. In addition, an "expiry"
        property is added to the Item properties indicating the earliest
        expiry time for any assets that were signed.

    """
    if copy:
        item = item.clone()
    urls = [asset.href for asset in item.assets.values()]
    signed_urls = sign_urls(urls=urls)
    for key, asset in item.assets.items():
        item.assets[key].href = signed_urls[asset.href]
    return item


@sign.register(Asset)
def sign_asset(asset: Asset, copy: bool = True) -> Asset:
    """Sign a PySTAC asset.

    Args:
        asset (Asset): The Asset to sign
        copy (bool): Whether to copy (clone) the asset or mutate it inplace.

    Returns:
        Asset: An asset where the HREF is replaced with a
        signed version.

    """
    if copy:
        asset = asset.clone()
    asset.href = sign_urls([asset.href])[asset.href]
    return asset


@sign.register(ItemCollection)
def sign_item_collection(
    item_collection: ItemCollection, copy: bool = True
) -> ItemCollection:
    """Sign a PySTAC item collection.

    Args:
        item_collection (ItemCollection): The ItemCollection whose assets will
            be signed
        copy (bool): Whether to copy (clone) the ItemCollection or mutate it
            inplace.

    Returns:
        ItemCollection: An ItemCollection where all assets'
        HREFs for each item have been replaced with a signed version. In
        addition, an "expiry" property is added to the Item properties
        indicating the earliest expiry time for any assets that were signed.

    """
    if copy:
        item_collection = item_collection.clone()
    urls = [asset.href for item in item_collection for asset in item.assets.values()]
    signed_urls = sign_urls(urls=urls)
    for item in item_collection:
        for key, asset in item.assets.items():
            item.assets[key].href = signed_urls[asset.href]
    return item_collection


@sign.register(ItemSearch)
def _search_and_sign(search: ItemSearch, copy: bool = True) -> ItemCollection:
    """Perform a PySTAC Client search, and sign the resulting item collection.

    Args:
        search (ItemSearch): The ItemSearch whose resulting item assets will
            be signed
        copy (bool): No effect.

    Returns:
        ItemCollection: The resulting ItemCollection of the search where all
            assets' HREFs for each item have been replaced with a signed
            version. In addition, a "expiry" property is added to the Item
            properties indicating the earliest expiry time for any assets that
            were signed.

    """
    if pystac_client.__version__ >= "0.5.0":
        items = search.item_collection()
    else:
        items = search.get_all_items()
    return sign(items)


@sign.register(Collection)
def sign_collection(collection: Collection, copy: bool = True) -> Collection:
    """Sign a collection.

    Args:
        collection: STAC Collection
        copy: copy or not the input

    Returns:
        signed (Collection): the STAC collection, now with signed URLs.

    """
    if copy:
        # https://github.com/stac-utils/pystac/pull/834 fixed asset dropping
        assets = collection.assets
        collection = collection.clone()
        if assets and not collection.assets:
            collection.assets = deepcopy(assets)

    urls = [collection.assets[key].href for key in collection.assets]
    signed_urls = sign_urls(urls=urls)
    for key, asset in collection.assets.items():
        collection.assets[key].href = signed_urls[asset.href]
    return collection


@sign.register(collections.abc.Mapping)
def sign_mapping(mapping: Mapping, copy: bool = True) -> Mapping:
    """Sign a mapping.

    Args:
        mapping (Mapping):

        The mapping (e.g. dictionary) to sign. This method can sign

            * Kerchunk-style references, which signs all URLs under the
              ``templates`` key. See https://fsspec.github.io/kerchunk/
              for more.
            * STAC items
            * STAC collections
            * STAC ItemCollections

        copy: Whether to copy (clone) the mapping or mutate it inplace.

    Returns:
        signed (Mapping): The dictionary, now with signed URLs.

    """
    if copy:
        mapping = deepcopy(mapping)

    types = (STACObjectType.ITEM, STACObjectType.COLLECTION)
    if all(key in mapping for key in ["version", "templates", "refs"]):
        urls = list(mapping["templates"].values())
        signed_urls = sign_urls(urls=urls)
        for key, url in mapping["templates"].items():
            mapping["templates"][key] = signed_urls[url]

    elif identify_stac_object_type(cast(Dict[str, Any], mapping)) in types:
        urls = [val["href"] for val in mapping["assets"].values()]
        signed_urls = sign_urls(urls=urls)
        for val in mapping["assets"].values():
            url = val["href"]
            val["href"] = signed_urls[url]

    elif mapping.get("type") == "FeatureCollection" and mapping.get("features"):
        urls = [
            val["href"]
            for feat in mapping["features"]
            for val in feat.get("assets", {}).values()
        ]
        signed_urls = sign_urls(urls=urls)
        for feature in mapping["features"]:
            for val in feature.get("assets", {}).values():
                url = val["href"]
                val["href"] = signed_urls[url]

    return mapping


sign_reference_file = sign_mapping


def _generic_get_signed_urls(
    urls: List[str],
    route: SignURLRoute,
) -> Dict[str, SignedURL]:
    """Get multiple signed URLs.

    This will use the URL from the cache if it's present and not too close
    to expiring. The generated URL will be placed in the cache.

    Args:
        urls: urls
        route: route (API)

    Returns:
        SignedURL: the signed URL

    """
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    log.debug("Get signed URLs for %s", urls)
    start_time = time.time()

    signed_urls = {}
    for url in urls:
        signed_url_in_cache = CACHE.get(url)
        if signed_url_in_cache and route == SignURLRoute.SIGN_URLS_GET:
            log.debug("URL %s already in cache", url)
            ttl = signed_url_in_cache.ttl()
            log.debug("URL %s TTL is %s", url, ttl)
            if ttl > ENV.tld_ttl_margin:
                log.debug(
                    "Using cache (%s > %s)",
                    ttl,
                    ENV.tld_ttl_margin,
                )
                signed_urls[url] = signed_url_in_cache
    not_signed_urls = [url for url in urls if url not in signed_urls]
    log.debug("Already signed URLs:\n %s", signed_urls)
    log.debug("Not signed URLs:\n %s", not_signed_urls)

    if not_signed_urls:
        # Refresh the token if there's less than
        # `settings.teledetection_ttl_margin seconds` remaining, in order to
        # give a small amount of time to do stuff with the url
        n_urls = len(not_signed_urls)
        log.debug("Number of URLs to sign: %s", n_urls)
        n_chunks = math.ceil(n_urls / MAX_URLS)
        log.debug("Number of chunks of URLs to sign: %s", n_chunks)
        for i_chunk in range(n_chunks):
            log.debug("Processing chunk %s/%s", i_chunk + 1, n_chunks)
            chunk_start = i_chunk * MAX_URLS
            chunk_end = min(chunk_start + MAX_URLS, n_urls)
            not_signed_urls_chunk = not_signed_urls[chunk_start:chunk_end]
            params: Dict[str, Any] = {"urls": not_signed_urls_chunk}
            if ENV.tld_url_duration:
                params["duration_seconds"] = ENV.tld_url_duration
            response = session.post(route=route.value, params=params)
            signed_url_batch = SignedURLBatch(**response.json())
            if not signed_url_batch:
                raise ValueError(
                    f"No signed url batch found in response: {response.json()}"
                )
            if not all(key in signed_url_batch.hrefs for key in not_signed_urls_chunk):
                raise ValueError(
                    f"URLs to sign are {not_signed_urls_chunk} but returned "
                    f"signed URLs"
                    f"are for {signed_url_batch.hrefs.keys()}"
                )
            for url, href in signed_url_batch.hrefs.items():
                signed_url = SignedURL(expiry=signed_url_batch.expiry, href=href)
                if route == SignURLRoute.SIGN_URLS_GET:
                    # Only put GET urls in cache
                    CACHE[url] = signed_url
                signed_urls[url] = signed_url
        log.debug(
            "Got signed urls %s in %s seconds",
            signed_urls,
            f"{time.time() - start_time:.2f}",
        )

    return signed_urls



"""An example showing how to create a tiny STAC collection with an item."""

from datetime import datetime
import pystac


# Create a collection
col1 = pystac.Collection(
    id="collection-test1",
    extent=pystac.Extent(
        pystac.SpatialExtent(4 * [0]),
        pystac.TemporalExtent(intervals=[[None, None]]),
    ),
    description="Some collection for tests",
    title="Collection test 1",
)

# Create an item
item1 = pystac.Item(
    id="item_23102024",
    datetime=datetime(year=2024, month=10, day=23),
    bbox=[2.49, 46.59, 3.26, 47.14],
    geometry={
        "type": "Polygon",
        "coordinates": [
            [[2.49, 47.14], [3.26, 47.13], [3.25, 46.59], [2.49, 46.59], [2.49, 47.14]]
        ],
    },
    assets={
        "estim": pystac.Asset(href="/tmp/estim.tif", media_type=pystac.MediaType.COG),
        "conf": pystac.Asset(href="/tmp/conf.tif", media_type=pystac.MediaType.COG),
    },
    properties={},
)

# Append item1 to collection
col1.add_item(item1)

# Update the collection spatial/temporal extent
col1.update_extent_from_items()

# Save the collection
col1.normalize_hrefs("/tmp/collection-test1")
col1.save(pystac.CatalogType.ABSOLUTE_PUBLISHED)



"""STAC stuff."""

import os
import re
import shutil
import json
from dataclasses import dataclass
from typing import List, cast
from urllib.parse import urljoin

import pystac
import pystac_client
import requests
from requests.exceptions import HTTPError
from requests.adapters import HTTPAdapter, Retry
from pystac import Collection, Item, ItemCollection
from rich.pretty import pretty_repr

from teledetection.sdk.settings import ENV
from teledetection.sdk.logger import get_logger_for
from teledetection.sdk.http import get_headers
from teledetection.sdk.signing import sign, sign_inplace
from .transfer import push
from . import raster

logger = get_logger_for(__name__)
TIMEOUT = ENV.tld_request_timeout

DEFAULT_STAC_EP = "https://api.stac.teledetection.fr"
DEFAULT_S3_EP = "https://s3-data.meso.umontpellier.fr"
DEFAULT_S3_STORAGE = "sm1-gdc-ext"


class STACObjectUnresolved(Exception):
    """Unresolved STAC object exception."""


class UnconsistentCollectionIDs(Exception):
    """Inconsistent STAC collection exception."""


class UnconsistentAssetNaming(Exception):
    """Inconsistent Asset Naming exception."""


class LogException(Exception):
    """Inconsistent Asset Naming exception."""


def _check_naming_is_compliant(s: str, allow_dot=False, allow_slash=False):
    _s = re.sub(r"[-|_]", r"", s)
    if allow_slash:
        _s = re.sub(r"\/", r"", _s)
    if allow_dot:
        _s = re.sub(r"\.", r"", _s)
    if not _s.isalnum():
        raise UnconsistentAssetNaming(
            f"{_s} does not only contain alphanumeric or - or _ chars"
        )


def create_session():
    """Create a requests session."""
    sess = requests.Session()
    retries = Retry(
        total=3,
        backoff_factor=1,
        status_forcelist=[
            408,
            419,
            425,
            500,
            502,
            503,
            504,
        ],
        allowed_methods=frozenset(["PUT", "POST"]),
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retries)
    sess.mount("http://", adapter=adapter)
    sess.mount("https://", adapter=adapter)
    return sess


def asset_exists(asset_url: str) -> bool:
    """Check that the item provided in parameter exists and is accessible."""
    sess = create_session()
    asset_url_signed = sign(asset_url)
    res = sess.get(asset_url_signed, stream=True)
    if res.status_code == 200:
        logger.info("Asset %s already exists.", asset_url)
        return True
    return False


def post_or_put(url: str, data: dict):
    """Post or put data to url."""
    headers = get_headers()
    sess = create_session()

    resp = sess.post(url, json=data, headers=headers, timeout=TIMEOUT)

    if resp.status_code == 409:
        # Exists, so update
        logger.info("Item at %s already exists, doing a PUT", url)
        resp = sess.put(
            f"{url}/{data['id']}",
            json=data,
            headers=headers,
            timeout=TIMEOUT,
        )
        # Unchanged may throw a 404
        if not resp.status_code == 404:
            resp.raise_for_status()

    try:
        resp.raise_for_status()
    except HTTPError as e:
        try:
            logger.error("Server returned: %s", pretty_repr(resp.json()))
        except LogException:
            logger.error("Server returned: %s", resp.text)
        raise e


def load_stac_obj(obj_pth: str) -> Collection | ItemCollection | Item:
    """Load a STAC object serialized on disk."""
    for obj_name, cls in {
        "collection": Collection,
        "item collection": ItemCollection,
        "item": Item,
    }.items():
        logger.debug("Try to read file %s", obj_pth)
        try:
            obj = getattr(cls, "from_file")(obj_pth)
            logger.info("Loaded %s from file %s", obj_name, obj_pth)
            logger.debug(obj.to_dict())
            return obj
        except pystac.errors.STACTypeError:
            pass

    raise STACObjectUnresolved(f"Cannot resolve STAC object ({obj_pth})")


def get_assets_root_dir(items: List[Item]) -> str:
    """Get the common prefix of all items assets paths.

    If the the common prefix is not a folder (/tmp/test1/a.tif, /tmp/test2/b.tif), returns /tmp.
    """
    prefix = os.path.commonpath(
        [asset.href for item in items for asset in item.assets.values()]
    )
    if os.path.isdir(prefix):
        return prefix + "/"
    return os.path.dirname(prefix) + "/"


def check_items_col_id(items: List[Item]):
    """Check that items have the same col_id."""
    if len(set(item.collection_id for item in items)) > 1:
        raise UnconsistentCollectionIDs("Collection ID must be the same for all items!")


def get_col_href(col: Collection):
    """Retrieve collection href."""
    for link in col.links:
        if link.rel == "self":
            return link.href
    return ""


def get_col_items(col: Collection) -> List[Item]:
    """Retrieve collection items."""
    col_href = get_col_href(col=col)
    return [
        cast(
            Item,
            load_stac_obj(
                os.path.join(os.path.dirname(col_href), link.href[2:])
                if link.href.startswith("./")
                else link.href
            ),
        )
        for link in col.links
        if link.rel == "item"
    ]


@dataclass
class StacTransactionsHandler:
    """Handle STAC and storage transactions."""

    stac_endpoint: str = DEFAULT_STAC_EP
    sign: bool = False
    """Sign URLs when fetching remote items."""

    @property
    def client(self):
        """STAC API client."""
        return pystac_client.Client.open(
            self.stac_endpoint,
            modifier=sign_inplace if self.sign else None,
        )

    def delete_item_or_col(self, col_id: str, item_id: str = ""):
        """Delete an item or a collection."""
        logger.info("Deleting %s%s", col_id, f"/{item_id}" if item_id else "")
        if item_id:
            url = f"{self.stac_endpoint}/collections/{col_id}/items/{item_id}"
        else:
            url = f"{self.stac_endpoint}/collections/{col_id}"
        resp = requests.delete(
            url,
            headers=get_headers(),
            timeout=TIMEOUT,
        )
        if resp.status_code != 200:
            logger.warning("Deletion failed (%s)", resp.text)

    def get_items(self, col_id: str, max_items: int = 10):
        """Get items in a collection."""
        logger.info("Get collection %s items", col_id)
        res = self.client.search(collections=[col_id], max_items=max_items)
        return list(res.items())

    def get_item(self, col_id: str, item_id: str) -> Item:
        """Retrieve a remote item."""
        logger.info("Retrieve item %s from collection %s", item_id, col_id)
        col = self.client.get_collection(col_id)
        if not col:
            raise ValueError(f"Collection {col_id} not found")
        item = col.get_item(item_id)
        if not item:
            raise ValueError(f"Item {item_id} (from collection {col_id}) not found")
        return item

    def publish_collection(self, col: Collection):
        """Publish an empty collection."""
        _check_naming_is_compliant(col.id)
        logger.info('Publishing collection "%s"', col.id)
        post_or_put(url=urljoin(self.stac_endpoint, "/collections"), data=col.to_dict())

    def publish_item(self, item: Item):
        """Publish an item."""
        _check_naming_is_compliant(item.id)
        col_id = item.collection_id
        logger.info('Publishing item "%s" in collection "%s"', item.id, col_id)
        post_or_put(
            urljoin(self.stac_endpoint, f"collections/{col_id}/items"),
            item.to_dict(transform_hrefs=False),
        )

    def update_collection_extent(self, col_id: str):
        """Update collection extent."""
        logger.info("Updating collection extent")
        results = self.client.search(limit=1000, collections=[col_id])
        col = self.client.get_collection(col_id)
        col.extent = pystac.Extent.from_items(items=list(results.items()))
        col.clear_links("items")
        self.publish_collection(col=col)

    def load_and_save(
        self, col_id: str, obj_pth: str, item_id: str = "", pretty: bool = True
    ):
        """Load and save locally (as .json file) the remote STAC object."""
        obj = (
            self.get_item(col_id=col_id, item_id=item_id)
            if item_id
            else self.client.get_collection(col_id)
        )
        logger.info("Writing file %s", obj_pth)
        with open(obj_pth, "w", encoding="utf-8") as file:
            json.dump(obj.to_dict(), file, indent=2 if pretty else None)

    def load_and_publish(self, obj_pth: str):
        """Load and publish the serialized STAC object."""
        obj = load_stac_obj(obj_pth=obj_pth)
        if isinstance(obj, Item):
            self.publish_item(item=obj)
        elif isinstance(obj, Collection):
            self.publish_collection(col=obj)
        elif isinstance(obj, ItemCollection):
            for item in obj.items:
                self.publish_item(item=item)
        else:
            raise TypeError(
                f"Invalid type, must be ItemCollection or Collection (got {type(obj)})"
            )


@dataclass
class StacUploadTransactionsHandler(StacTransactionsHandler):
    """Handle STAC and storage transactions."""

    storage_endpoint: str = DEFAULT_S3_EP
    storage_bucket: str = DEFAULT_S3_STORAGE
    assets_overwrite: bool = False
    """Overwrite assets on S3 if already existing."""
    keep_cog_dir: str = ""

    def publish_item_and_push_assets(self, item: Item, assets_root_dir: str):
        """Publish an item and push all its assets.

        Args:
            item: Stac item to publish
            assets_root_dir: Common path to all files, defined as assets root dir
        """
        tgt_root_url = urljoin(
            self.storage_endpoint, f"{self.storage_bucket}/{item.collection_id}/"
        )

        logger.debug("Itemid = %s", item.id)

        _check_naming_is_compliant(self.storage_bucket)
        _check_naming_is_compliant(item.id)
        for _, asset in item.assets.items():
            local_filename = asset.href
            logger.debug("Local file: %s", local_filename)

            file_relative_path = os.path.relpath(local_filename, assets_root_dir)
            target_url = urljoin(tgt_root_url, file_relative_path)

            # Check that url part after storage bucket is compliant
            _check_naming_is_compliant(
                file_relative_path,
                allow_dot=True,
                allow_slash=True,
            )
            logger.debug("Target file: %s", target_url)

            # Add raster metadata to asset
            logger.debug("Updating assets metadata for rasters...")
            if raster.is_raster(local_filename):
                raster.apply_proj_extension(asset)
                raster.apply_raster_extension(asset)
                asset.media_type = pystac.MediaType.COG

            # Skip when target file exists and overwrite is not enabled
            if not self.assets_overwrite:
                if asset_exists(target_url):
                    asset.href = target_url
                    continue

            # Check is_cog, converts if not
            cogconv = False
            if raster.is_raster(local_filename):
                if not raster.is_cog(local_filename):
                    orig_filename = local_filename
                    local_filename = raster.convert_to_cog(
                        orig_filename,
                        keep_cog_dir=self.keep_cog_dir,
                    )
                    cogconv = True

            # Upload file
            logger.info("Uploading %s to %s...", local_filename, target_url)
            try:
                push(local_filename=local_filename, target_url=target_url)
            except Exception as e:
                logger.error(e)
                raise e

            # Update assets hrefs
            logger.debug("Updating assets HREFs ...")
            asset.href = target_url

            # Delete temp cog
            if cogconv and not self.keep_cog_dir:
                logger.debug("Deleting temporary COG ...")
                shutil.rmtree(os.path.dirname(local_filename))
                local_filename = orig_filename

        # Add published metadata to item
        logger.debug("Updating item metadata ...")
        raster.apply_published_extension(item)

        # Push item
        self.publish_item(item=item)

    def publish_items_and_push_assets(self, items: List[Item]):
        """Publish items."""
        if not items:
            logger.info("No item to publish.")
            return
        check_items_col_id(items=items)
        assets_root_dir = get_assets_root_dir(items=items)
        logger.debug("Assets root directory: %s", assets_root_dir)
        for item in items:
            self.publish_item_and_push_assets(
                item=item, assets_root_dir=assets_root_dir
            )
        # Update collection extent
        col_id = items[0].collection_id
        if not col_id:
            raise UnconsistentCollectionIDs(
                f"Collection id is None for item {items[0].id}"
            )
        self.update_collection_extent(col_id=col_id)

    def publish_collection_with_items(self, col: Collection):
        """Publish a collection and all its items."""
        items = get_col_items(col=col)
        check_items_col_id(items)
        self.publish_collection(col=col)
        self.publish_items_and_push_assets(items=items)

    def publish_item_collection(self, item_collection: ItemCollection):
        """Publish an item collection and all of its items."""
        self.publish_items_and_push_assets(items=item_collection.items)

    def load_and_publish(self, obj_pth: str):
        """Load and publish the serialized STAC object."""
        obj = load_stac_obj(obj_pth=obj_pth)
        if isinstance(obj, Item):
            self.publish_items_and_push_assets(items=[obj])
        elif isinstance(obj, Collection):
            self.publish_collection_with_items(col=obj)
        elif isinstance(obj, ItemCollection):
            self.publish_item_collection(item_collection=obj)
        else:
            raise TypeError(
                f"Invalid type, must be ItemCollection or Collection (got {type(obj)})"
            )



"""Logger module."""

import os
import logging

# Logger
LOGLEVEL = os.environ.get("LOGLEVEL") or "INFO"
logging.basicConfig(level=LOGLEVEL)


def get_logger_for(name: str):
    """Get logger for a named module."""
    logger = logging.getLogger(name)
    logger.setLevel(level=LOGLEVEL)
    return logger
